from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from pydantic import validator
from sqlmodel import SQLModel

from ._validators import valint
from ._validators import valstr
from .task import TaskExport
from .task import TaskImport
from .task import TaskRead


__all__ = (
    "WorkflowCreate",
    "WorkflowRead",
    "WorkflowUpdate",
    "WorkflowImport",
    "WorkflowExport",
    "WorkflowTaskCreate",
    "WorkflowTaskImport",
    "WorkflowTaskExport",
    "WorkflowTaskRead",
    "WorkflowTaskUpdate",
)


class _WorkflowTaskBase(SQLModel):

    meta: Optional[Dict[str, Any]] = None
    args: Optional[Dict[str, Any]] = None


class WorkflowTaskCreate(_WorkflowTaskBase):
    order: Optional[int]
    task_id: int
    # Validators
    _order = validator("order", allow_reuse=True)(valint("order", min_val=0))
    _task_id = validator("task_id", allow_reuse=True)(valint("task_id"))


class WorkflowTaskRead(_WorkflowTaskBase):
    id: int
    order: Optional[int]
    workflow_id: int
    task_id: int
    task: TaskRead


class WorkflowTaskImport(_WorkflowTaskBase):
    task: TaskImport


class WorkflowTaskExport(_WorkflowTaskBase):
    task: TaskExport


class WorkflowTaskUpdate(_WorkflowTaskBase):
    # Validators
    @validator("meta")
    def check_no_parallelisation_level(cls, m):
        if "parallelization_level" in m:
            raise ValueError(
                "Overriding task parallelization level currently not allowed"
            )
        return m


class _WorkflowBase(SQLModel):
    name: str


class WorkflowRead(_WorkflowBase):
    id: int
    project_id: int
    task_list: List[WorkflowTaskRead]


class WorkflowCreate(_WorkflowBase):
    project_id: int

    # Validators
    _name = validator("name", allow_reuse=True)(valstr("name"))
    _project_id = validator("project_id", allow_reuse=True)(
        valint("project_id")
    )


class WorkflowUpdate(_WorkflowBase):
    name: Optional[str]
    reordered_workflowtask_ids: Optional[List[int]]

    # Validators
    _name = validator("name", allow_reuse=True)(valstr("name"))

    @validator("reordered_workflowtask_ids")
    def check_positive_and_unique(cls, value):
        if any(i < 0 for i in value):
            raise ValueError("Negative `id` in `reordered_workflowtask_ids`")
        if len(value) != len(set(value)):
            raise ValueError("`reordered_workflowtask_ids` has repetitions")
        return value


class WorkflowImport(_WorkflowBase):
    task_list: List[WorkflowTaskImport]

    # Validators
    _name = validator("name", allow_reuse=True)(valstr("name"))


class WorkflowExport(_WorkflowBase):
    task_list: List[WorkflowTaskExport]



import pytest
from devtools import debug
from pydantic.error_wrappers import ValidationError

from schemas import ManifestV1
from schemas import TaskManifestV1


def test_ManifestV1():

    t = TaskManifestV1(
        name="task",
        executable="executable",
        input_type="input_type",
        output_type="output_type",
        default_args={"arg": "val"},
    )
    m = ManifestV1(manifest_version="1", task_list=[t])
    debug(m)

    with pytest.raises(ValidationError):
        ManifestV1(manifest_version="2", task_list=[])



import pytest
from devtools import debug
from pydantic.error_wrappers import ValidationError

from schemas import TaskCreate
from schemas import TaskUpdate


def test_task_update():
    # Successful creation, with many unset fields
    t = TaskUpdate(name="name")
    debug(t)
    assert list(t.dict(exclude_none=True).keys()) == ["name"]
    assert list(t.dict(exclude_unset=True).keys()) == ["name"]


def test_task_create():
    # Successful creation
    t = TaskCreate(
        name="task",
        source="source",
        command="command",
        input_type="input_type",
        output_type="output_type",
    )
    debug(t)
    # Missing arguments
    with pytest.raises(ValidationError):
        TaskCreate(name="task", source="source")



import json
from typing import Optional

from ..authclient import AuthClient
from ..common.schemas import StateRead
from ..common.schemas import TaskCollectPip
from ..common.schemas import TaskCreate
from ..common.schemas import TaskRead
from ..common.schemas import TaskUpdate
from ..config import settings
from ..interface import BaseInterface
from ..interface import PrintInterface
from ..interface import RichJsonInterface
from ..response import check_response
from .utils import get_cached_task_by_name
from .utils import refresh_task_cache


async def task_list(client: AuthClient, **kwargs) -> RichJsonInterface:
    task_list = await refresh_task_cache(client=client, **kwargs)
    return RichJsonInterface(retcode=0, data=task_list)


async def task_collect_pip(
    client: AuthClient,
    *,
    package: str,
    package_version: Optional[str] = None,
    python_version: Optional[str],
    package_extras: Optional[str] = None,
    private: Optional[bool] = False,
    batch: bool = False,
    **kwargs,
) -> BaseInterface:
    task_collect = TaskCollectPip(
        package=package,
        version=package_version,
        python_version=python_version,
        package_extras=package_extras,
    )

    res = await client.post(
        f"{settings.BASE_URL}/task/collect/pip/?public={not private}",
        json=task_collect.dict(),
    )

    state = check_response(
        res, expected_status_code=[200, 201], coerce=StateRead
    )
    if batch:
        output = f"{state.id} {state.data['venv_path']}"
        return PrintInterface(retcode=0, data=output)
    else:
        return RichJsonInterface(retcode=0, data=state.sanitised_dict())


async def task_collection_check(
    client: AuthClient,
    *,
    state_id: int,
    include_logs: bool,
    do_not_separate_logs: bool = False,
    **kwargs,
) -> BaseInterface:

    res = await client.get(
        f"{settings.BASE_URL}/task/collect/{state_id}?verbose={include_logs}"
    )
    state = check_response(res, expected_status_code=200, coerce=StateRead)

    state_dict = state.sanitised_dict()

    # Remove key-value pairs with None value
    state_dict["data"] = {
        key: val for (key, val) in state_dict["data"].items() if val
    }

    if (not include_logs) or do_not_separate_logs:
        return RichJsonInterface(retcode=0, data=state_dict)
    else:
        log = state_dict["data"].pop("log")
        extra_lines = ["\nThis is the task-collection log:\n", log]
        return RichJsonInterface(
            retcode=0, data=state_dict, extra_lines=extra_lines
        )


async def task_new(
    client: AuthClient,
    *,
    name: str,
    command: str,
    source: str,
    batch: bool = False,
    input_type: Optional[str] = "Any",
    output_type: Optional[str] = "Any",
    default_args_file: Optional[str] = None,
    meta_file: Optional[str] = None,
    **kwargs,
) -> BaseInterface:
    optionals = {}
    if default_args_file:
        with open(default_args_file, "r") as f:
            optionals["default_args"] = json.load(f)
    if meta_file:
        with open(meta_file, "r") as f:
            optionals["meta"] = json.load(f)
    payload = TaskCreate(
        name=name,
        command=command,
        source=source,
        input_type=input_type,
        output_type=output_type,
        **optionals,
    ).dict(exclude_unset=True)
    res = await client.post(f"{settings.BASE_URL}/task/", json=payload)
    new_task = check_response(res, expected_status_code=201, coerce=TaskRead)

    if batch:
        return PrintInterface(retcode=0, data=str(new_task.id))
    else:
        return RichJsonInterface(retcode=0, data=new_task.dict())


async def task_edit(
    client: AuthClient,
    *,
    task_id_or_name: str,
    name: Optional[str] = None,
    command: Optional[str] = None,
    input_type: Optional[str] = None,
    output_type: Optional[str] = None,
    default_args_file: Optional[str] = None,
    meta_file: Optional[str] = None,
    **kwargs,
) -> BaseInterface:
    update = {}
    if name:
        update["name"] = name
    if command:
        update["command"] = command
    if input_type:
        update["input_type"] = input_type
    if output_type:
        update["output_type"] = output_type
    if default_args_file:
        with open(default_args_file, "r") as f:
            update["default_args"] = json.load(f)
    if meta_file:
        with open(meta_file, "r") as f:
            update["meta"] = json.load(f)

    task_update = TaskUpdate(**update)  # validation
    payload = task_update.dict(exclude_unset=True)
    if not payload:
        return PrintInterface(retcode=1, data="Nothing to update")

    try:
        task_id = int(task_id_or_name)
    except ValueError:
        task_id = await get_cached_task_by_name(
            name=task_id_or_name, client=client
        )

    res = await client.patch(
        f"{settings.BASE_URL}/task/{task_id}", json=payload
    )
    new_task = check_response(res, expected_status_code=200, coerce=TaskRead)
    return RichJsonInterface(retcode=0, data=new_task.dict())


async def task_delete(
    client: AuthClient, task_id: int, **kwargs
) -> PrintInterface:

    raise NotImplementedError("task_delete")

    # res = await client.delete(
    #    f"{settings.BASE_URL}/xxxx"
    # )
    # check_response(res, expected_status_code=204)
    # return PrintInterface(retcode=0, data="")



import pytest
from devtools import debug
from pydantic.error_wrappers import ValidationError

from schemas import DatasetCreate
from schemas import DatasetRead
from schemas import DatasetUpdate
from schemas import ResourceRead


def test_dataset_create():
    # Successful creation
    d = DatasetCreate(name="name")
    # Successful sanification of whitespaces
    NAME = "name"
    d = DatasetCreate(name=f"   {NAME}   ")
    assert d.name == NAME
    assert not d.read_only  # Because of default False value
    # Missing argument
    with pytest.raises(ValidationError):
        d = DatasetCreate()
    # Empty-string argument
    with pytest.raises(ValidationError):
        d = DatasetCreate(name="  ")


def test_dataset_read():
    # Successful creation - empty resource_list
    d = DatasetRead(
        id=1, project_id=1, resource_list=[], name="n", read_only=True
    )
    debug(d)
    # Successful creation - non-trivial resource_list
    r1 = ResourceRead(id=1, dataset_id=1, path="/something")
    r2 = ResourceRead(id=1, dataset_id=1, path="/something")
    rlist = [r1, r2]
    d = DatasetRead(
        id=1, project_id=1, resource_list=rlist, name="n", read_only=False
    )
    debug(d)


def test_dataset_update():
    # Sanity check: attributes which are not set explicitly are not listed when
    # exclude_unset=True

    payload = dict(name="name")
    dataset_update_dict = DatasetUpdate(**payload).dict(exclude_unset=True)
    debug(dataset_update_dict)
    assert dataset_update_dict.keys() == payload.keys()

    payload = dict(type="type")
    dataset_update_dict = DatasetUpdate(**payload).dict(exclude_unset=True)
    debug(dataset_update_dict)
    assert dataset_update_dict.keys() == payload.keys()

    payload = dict(read_only=True)
    dataset_update_dict = DatasetUpdate(**payload).dict(exclude_unset=True)
    debug(dataset_update_dict)
    assert dataset_update_dict.keys() == payload.keys()

    payload = dict(read_only=True, name="name")
    dataset_update_dict = DatasetUpdate(**payload).dict(exclude_unset=True)
    debug(dataset_update_dict)
    assert dataset_update_dict.keys() == payload.keys()



import pytest
from devtools import debug
from pydantic.error_wrappers import ValidationError

from schemas import ApplyWorkflowCreate


def test_apply_workflow_create():
    # Valid ApplyWorkflowCreate instance / without output_dataset_id
    valid_args = dict(
        project_id=1,
        input_dataset_id=1,
        workflow_id=3,
        overwrite_input=True,
        worker_init="WORKER INIT",
    )
    job = ApplyWorkflowCreate(**valid_args)
    debug(job)
    assert job.output_dataset_id is None
    # Valid ApplyWorkflowCreate instance / with output_dataset_id
    valid_args["output_dataset_id"] = 3
    job = ApplyWorkflowCreate(**valid_args)
    debug(job)
    assert job.output_dataset_id
    # Multiple cases of invalid arguments
    for key in [
        "project_id",
        "input_dataset_id",
        "output_dataset_id",
        "workflow_id",
    ]:
        invalid_args = valid_args.copy()
        invalid_args[key] = -1
        debug(invalid_args)
        with pytest.raises(ValidationError):
            ApplyWorkflowCreate(**invalid_args)



import pytest
from devtools import debug
from pydantic.error_wrappers import ValidationError

from schemas import TaskImport
from schemas import TaskRead
from schemas import WorkflowCreate
from schemas import WorkflowImport
from schemas import WorkflowRead
from schemas import WorkflowTaskCreate
from schemas import WorkflowTaskImport
from schemas import WorkflowTaskRead
from schemas import WorkflowTaskUpdate
from schemas import WorkflowUpdate


def test_workflow_task_create():
    # Successful creation
    t = WorkflowTaskCreate(task_id=1)
    debug(t)
    # Missing arguments
    with pytest.raises(ValidationError):
        WorkflowTaskCreate(task_id=-1)
    # Several values of order
    t = WorkflowTaskCreate(task_id=1, order=1)
    t = WorkflowTaskCreate(task_id=1, order=0)
    with pytest.raises(ValidationError):
        WorkflowTaskCreate(task_id=1, order=-1)
    with pytest.raises(ValidationError):
        WorkflowTaskCreate(task_id=1, order=None)


def test_workflow_task_update():
    # Successful creation
    t = WorkflowTaskUpdate(meta=dict(something="else"))
    # Forbidden key-value update
    with pytest.raises(ValidationError):
        t = WorkflowTaskUpdate(meta=dict(parallelization_level="new"))
    debug(t)


def test_workflow_create():
    w = WorkflowCreate(name="workflow", project_id=1)
    debug(w)


def test_workflow_import():
    # Successful creation
    t = TaskImport(name="name", source="source")
    wft = WorkflowTaskImport(task=t)
    w = WorkflowImport(name="workflow", task_list=[wft])
    debug(w)
    # Empty-string argument
    with pytest.raises(ValidationError):
        WorkflowImport(name=" ", task_list=[wft])


def test_workflow_read_empty_task_list():
    w = WorkflowRead(id=1, name="workflow", project_id=1, task_list=[])
    debug(w)


def test_workflow_read_non_empty_task_list():
    # Create a TaskRead
    t1 = TaskRead(
        id=9,
        name="name",
        source="source",
        command="command",
        input_type="input_type",
        output_type="output_type",
        meta=dict(something="else"),
    )
    # Create two WorkflowTaskRead
    wft1 = WorkflowTaskRead(id=1, task_id=1, workflow_id=1, task=t1)
    wft2 = WorkflowTaskRead(id=2, task_id=1, workflow_id=1, task=t1)
    # Create a WorkflowRead
    w = WorkflowRead(
        id=1, name="workflow", project_id=1, task_list=[wft1, wft2]
    )
    debug(w)


def test_workflow_update():
    WorkflowUpdate(name="workflow", reordered_workflowtask_ids=[0, 1, 3, 2])
    WorkflowUpdate(name="workflow")
    WorkflowUpdate(reordered_workflowtask_ids=[0, 1, 3, 2])
    with pytest.raises(ValidationError):
        WorkflowUpdate(name="workflow", reordered_workflowtask_ids=[1, 3, 1])
    with pytest.raises(ValidationError):
        WorkflowUpdate(name="workflow", reordered_workflowtask_ids=[1, 3, -1])



import pytest
from devtools import debug
from pydantic.error_wrappers import ValidationError

from schemas import ProjectCreate


def test_project_create():
    # Successful creation
    p = ProjectCreate(name="my project")
    debug(p)
    # Check that whitespaces are stripped from beginning/end of string
    NAME = "some project name"
    p = ProjectCreate(name=f"  {NAME}  ")
    debug(p)
    assert p.name == NAME
    # Fail due to empty string
    with pytest.raises(ValidationError):
        ProjectCreate(name="  ")



from datetime import datetime

from devtools import debug

from schemas import _StateBase
from schemas import StateRead


def test_state():
    s = _StateBase(data={"some": "thing"}, timestamp=datetime.now())
    debug(s)
    debug(s.sanitised_dict())
    assert isinstance(s.sanitised_dict()["timestamp"], str)


def test_state_read():
    s = StateRead(data={"some": "thing"}, timestamp=datetime.now())
    debug(s)
    assert s.id is None

    s = StateRead(data={"some": "thing"}, timestamp=datetime.now(), id=1)
    debug(s)
    assert s.id == 1



import pytest
from devtools import debug
from pydantic.error_wrappers import ValidationError

from schemas import UserCreate


def test_user_create():
    # Without slurm_user attribute
    u = UserCreate(email="a@b.c", password="asd")
    debug(u)
    assert u.slurm_user is None
    # With valid slurm_user attribute
    u = UserCreate(email="a@b.c", password="asd", slurm_user="slurm_user")
    debug(u)
    assert u.slurm_user
    # With invalid slurm_user attribute
    with pytest.raises(ValidationError):
        u = UserCreate(email="a@b.c", password="asd", slurm_user="  ")
    # With valid cache_dir
    CACHE_DIR = "/xxx"
    u = UserCreate(email="a@b.c", password="asd", cache_dir=f"{CACHE_DIR}   ")
    assert u.cache_dir == CACHE_DIR
    # With invalid cache_dir attribute
    with pytest.raises(ValidationError) as e:
        u = UserCreate(email="a@b.c", password="asd", cache_dir="  ")
    debug(e.value)
    assert "cannot be empty" in e.value.errors()[0]["msg"]
    # With invalid cache_dir attribute
    with pytest.raises(ValidationError) as e:
        u = UserCreate(email="a@b.c", password="asd", cache_dir="xxx")
    debug(e.value)
    assert "must be an absolute path" in e.value.errors()[0]["msg"]
